/*
*  C Implementation: main
*
* Description: 
*
*
* Author: Hagen Meyer <hagen@hcmeyer.de>, (C) 2009
*
* Copyright: See COPYING file that comes with this distribution
*
*/

/** \mainpage
 * Nothing here, yet…
 */

/** \file
 * \brief The main file
 */

#define DEBUG		0

#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

/*##########################################################################
	CONSTANT DEFINITIONS
###########################################################################*/

/// \brief Register for the motor PWM
#define MOTOR_SPEED	OCR0B

#define SPEED_NORMAL		65
#define SPEED_FAST		230
#define STOP_PROBABILITY	5

/// \name LED hardware connections
//@{
#define PORT_LED01	PORTA
#define PORT_LED23	PORTD
#define PIN_LED0	1
#define PIN_LED1	0
#define PIN_LED2	2
#define PIN_LED3	3
//@}

/// \name Remote-Control hardware connections
//@{
#define REMOTE_FAST	_BV(PD0) // RxD
#define REMOTE_SLOW	_BV(PD1) // TxD
#define REMOTE_STOP	(_BV(PD0)|_BV(PD1)) // both
//@}

/// \name LED commands
//@{
#define OFF		0
#define ON		1
#define TOGGLE		2
//@}

#define LED_ALL		4

/** \name DIP-switch settings
 * \brief The different modes selectable with the DIP switches
 */
//@{
#define GAME_MODE_SLOW		0 << 0 ///< DIP1..2 = 00 (booooring)
#define GAME_MODE_FAST		1 << 0 ///< DIP1..2 = 10 (not soo boring)
#define GAME_MODE_RANDOM	2 << 0 ///< DIP1..2 = 01 (magic :) )
#define GAME_MODE_RANDOM_STOP	3 << 0 ///< DIP1..2 = 11 (more magic :) )
#define LED_MODE_OFF		0 << 2 ///< DIP3..4 = 00
#define LED_MODE_ON		1 << 2 ///< DIP3..4 = 10
#define LED_MODE_RAND		2 << 2 ///< DIP3..4 = 01
#define REMOTE_MODE_25		0 << 4 ///< DIP5..6 = 00 
#define REMOTE_MODE_50		1 << 4 ///< DIP5..6 = 10
#define REMOTE_MODE_75		2 << 4 ///< DIP5..6 = 01
#define REMOTE_MODE_100		3 << 4 ///< DIP5..6 = 11
//@}

/*##########################################################################
	MACROS
###########################################################################*/

/// \brief reads the Game mode from the DIP-switches
#define getGameMode()		((~PINB) & 3)
/// \brief reads the LED mode from the DIP-switches
#define getLEDMode()		((~PINB) & (3<<2))
/// \brief reads the remote-control mode from the DIP-switches
#define getRemoteMode()		((~PINB) & (3<<4))
/// \brief reads the buttons of the remote control
#define getRemoteButtons()	((~PIND) & (_BV(PD0)|_BV(PD1)))
/// \brief reads the position of the ON/OFF switch (for the motor)
#define isOn()			!(PIND & _BV(PD4))
/// \brief starts Timer 1 with prescaler 1024 (~1ms/step)
#define runTimer()		TCCR1B = _BV(CS12) | _BV(CS10)
/// \brief resets Timer 1
#define resetTimer()		TCNT1 = 0
/// \brief stops Timer 1
#define stopTimer()		TCCR1B = 0
/// \brief set the Timer-value
#define setTimer(val)		TCNT1 = (val)
/// \brief gets the value of Timer 1 (elapsed time in ~ms)
#define time()			TCNT1

/*##########################################################################
	PROTOTYPES
###########################################################################*/

int main (void);
void switchLED(uint8_t led, uint8_t state);
void LEDMagic(void);
void GameMagic(void);

/*##########################################################################
	GLOBAL VARIABLES
###########################################################################*/

uint8_t random_nr;
uint8_t led_mode;
uint8_t remote_prev;		// previous state of the remote
uint8_t remote_accepted;	// remote commands which are accepted (selectable probability)
uint8_t speed;

/*##########################################################################
	SUBROUTINES
###########################################################################*/

/** \brief Routine to switch the LEDS
 *  \param led Which LED to switch (0 to 3, or 4 for all)
 *  \param state What to do (see \ref LED commands)
 *
 */
void switchLED(uint8_t led, uint8_t state){
	uint8_t *port;
	uint8_t operand;
	switch(led){
		case 0: port = (uint8_t*)&PORT_LED01; operand = _BV(PIN_LED0); break;
		case 1: port = (uint8_t*)&PORT_LED01; operand = _BV(PIN_LED1); break;
		case 2: port = (uint8_t*)&PORT_LED23; operand = _BV(PIN_LED2); break;
		case 3: port = (uint8_t*)&PORT_LED23; operand = _BV(PIN_LED3); break;
		case 4: switch(state){
			case ON: PORT_LED01 |= _BV(PIN_LED0) | _BV(PIN_LED1);
				PORT_LED23 |= _BV(PIN_LED2) | _BV(PIN_LED3);
				break;
			case OFF: PORT_LED01 &= ~(_BV(PIN_LED0) | _BV(PIN_LED1));
				PORT_LED23 &= ~(_BV(PIN_LED2) | _BV(PIN_LED3));
				break;
			case TOGGLE: PORT_LED01 ^= _BV(PIN_LED0) | _BV(PIN_LED1);
				PORT_LED23 ^= _BV(PIN_LED2) | _BV(PIN_LED3);
				break;
			}
			return;
		default: return;
	}
	switch(state){
		case ON: *port |= operand; break;
		case OFF: *port &= ~operand; break;
		case TOGGLE: *port ^= operand; break;
	}
}

/** \brief Function for the LED magic driven by the DIP-switches for LED mode
 */
void LEDMagic() {
	uint8_t tmp;
	///////
	// LED magic
	///////
	if(led_mode == 0) { // random mode
		if(time() % 256 == 0) switchLED(random_nr & 3,TOGGLE);
	} else if(led_mode == 1) { // circle 1
		tmp = time()/128;
		switchLED((tmp-1)&3,ON);
		switchLED(tmp&3,OFF);
	} else if(led_mode == 2) { // circle 2
		tmp = ~(time()/128);
		switchLED((tmp-1)&3,ON);
		switchLED(tmp&3,OFF);
	} else if(led_mode == 3) { // toggle
		if((time()/512)&1) switchLED(LED_ALL,TOGGLE);
	}
	///////
	// switch LED modes
	///////
	if(getLEDMode() == LED_MODE_ON) {
		switchLED(LED_ALL,ON);
		led_mode = 0xff;
	} else if(getLEDMode() == LED_MODE_OFF) {
		switchLED(LED_ALL,OFF);
		led_mode = 0xff;
	} else if((getLEDMode() == LED_MODE_RAND) && ((time()/1024) % 10 == 0)) { // random LED modes
		led_mode = random_nr & 3;
		if(led_mode == 3) {
			switchLED(LED_ALL,OFF);
			switchLED(0,ON);
			switchLED(2,ON);
		}
	}
}

/** \brief Function which drives the motor. Controlled by the DIP-switches for Game mode
 */
void GameMagic() {
	///////
	// Remote control
	///////
	uint8_t remote_curr = getRemoteButtons();
	if(remote_curr != remote_prev) {
		if((remote_curr == 0) || (getRemoteMode() == REMOTE_MODE_100)) {
			remote_accepted = remote_curr;
		} else if((getRemoteMode() == REMOTE_MODE_25) && ((random_nr & 3) == 0)) {
			remote_accepted = remote_curr;
		} else if((getRemoteMode() == REMOTE_MODE_50) && ((random_nr & 3) < 2)) {
			remote_accepted = remote_curr;
		} else if((getRemoteMode() == REMOTE_MODE_75) && ((random_nr & 3) < 3)) {
			remote_accepted = remote_curr;
		}
		remote_prev = remote_curr;
	}
	///////
	// Game magic
	///////
	if(getGameMode() == GAME_MODE_SLOW) { // boooring
		speed = SPEED_NORMAL;
	} else if(getGameMode() == GAME_MODE_FAST) { // not so boring
		speed = SPEED_FAST;
	} else { // magic :)
		if(time() % 256 == 0) {
			if ((random_nr >= SPEED_NORMAL) && (random_nr <= SPEED_FAST)) {
				speed = random_nr;
			} else if ((random_nr < SPEED_NORMAL) && (random_nr >= SPEED_NORMAL - 20)) { // higher probability for the lowest speed
			  	speed = SPEED_NORMAL;
			} else if ((random_nr > SPEED_FAST) && (random_nr <= SPEED_FAST + 10)) { // higher probability for the highest speed
				speed = SPEED_FAST;
			} else if (getGameMode() == GAME_MODE_RANDOM_STOP && random_nr < STOP_PROBABILITY) {
				speed = 0;
			}
		}
	}
	if(remote_accepted == REMOTE_FAST) {
		MOTOR_SPEED = SPEED_FAST;
	} else if(remote_accepted == REMOTE_SLOW) {
		MOTOR_SPEED = SPEED_NORMAL;
	} else if(remote_accepted == REMOTE_STOP) {
		MOTOR_SPEED = 0;
	} else {
		MOTOR_SPEED = speed;
	}
}

/*##########################################################################
	MAIN ROUTINE
###########################################################################*/

/** \brief The main function */
int main (void) {
/*##########################################################################
	INITIALIZATION
###########################################################################*/
	ACSR = _BV(ACD);		// disable analog comparator
	
	// LED0 and 1 as output
	DDRA = _BV(PA0) | _BV(PA1);

	// DIP-switches (PB0..5) as input w/ pullup
	PORTB = (uint8_t)(~(3<<6));

	// LED2 and 3 as output, Motor-PWM as output, ON/OFF switch as input w/ pullup, Remote-Switches as input w/ pullup
	DDRD = _BV(PD2) | _BV(PD3) | _BV(PD5);
	PORTD = _BV(PD0) | _BV(PD1) | _BV(PD4);

	MOTOR_SPEED = 0;
	// Non-Inverting 8 bit phase correct PWM (MAX=0xFF) on PD5, no prescaler
	TCCR0A = _BV(COM0B1) | _BV(WGM00);
	TCCR0B = _BV(CS00);

	extern unsigned short __heap_start;
	unsigned int seed = *((unsigned int *)__heap_start);
	srand(seed); 			// seed random number generator with uninitialized sram value
/*##########################################################################
	MAIN LOOP
###########################################################################*/
	led_mode = 0xff;
	remote_prev = 0;
	for(;;) {
		random_nr = (uint8_t)rand(); // number from 0 to 0xFF
		if(isOn()) {
			runTimer();
			if(time() < 500) { // start countdount
				if(getLEDMode() != LED_MODE_OFF) switchLED(0,ON);
				speed = SPEED_NORMAL;
				MOTOR_SPEED = speed;
			} else if(time() < 1000) {
				if(getLEDMode() != LED_MODE_OFF) switchLED(1,ON);
			} else if(time() < 1500) {
				if(getLEDMode() != LED_MODE_OFF) switchLED(2,ON);
			} else if(time() < 2000) {
				if(getLEDMode() != LED_MODE_OFF) switchLED(3,ON);
			} else if(time() > 65000) { // set timer back to avoid a new countdown
				setTimer(3000);
			} else if(time() >= 3000) { // start the magic ;)
				LEDMagic();
				GameMagic();
			}
		}
		else { // switch off everything
		  	MOTOR_SPEED = 0;
			switchLED(LED_ALL,OFF);
			stopTimer();
			resetTimer();
		}
	}
}
